#include <rt/event.h>

#include <rt/tick.h>

static inline bool match_any(uint32_t bits, uint32_t wait)
{
    return ((bits & wait) != UINT32_C(0));
}

static inline bool match_all(uint32_t bits, uint32_t wait)
{
    return ((bits & wait) == wait);
}

uint32_t rt_event_clear(struct rt_event *event, uint32_t clear)
{
    return rt_atomic_fetch_and(&event->bits, ~clear, RT_ATOMIC_ACQ_REL);
}

uint32_t rt_event_get(const struct rt_event *event)
{
    return rt_atomic_load(&event->bits, RT_ATOMIC_ACQUIRE);
}

uint32_t rt_event_set(struct rt_event *event, uint32_t set)
{
    const uint32_t oldbits =
        rt_atomic_fetch_or(&event->bits, set, RT_ATOMIC_RELEASE);
    rt_cond_broadcast(&event->cond);
    return oldbits;
}

uint32_t rt_event_wait(struct rt_event *event, uint32_t wait)
{
    uint32_t bits = rt_event_trywait(event, wait);
    if (match_any(bits, wait))
    {
        return bits;
    }
    rt_mutex_lock(&event->mutex);
    while (!match_any(bits = rt_event_trywait(event, wait), wait))
    {
        rt_cond_wait(&event->cond, &event->mutex);
    }
    rt_mutex_unlock(&event->mutex);
    return bits;
}

uint32_t rt_event_trywait(struct rt_event *event, uint32_t wait)
{
    uint32_t bits = rt_atomic_load(&event->bits, RT_ATOMIC_RELAXED);
    while (match_any(bits, wait) &&
           !rt_atomic_compare_exchange_weak(&event->bits, &bits, bits & ~wait,
                                            RT_ATOMIC_ACQUIRE,
                                            RT_ATOMIC_RELAXED))
    {
    }
    return bits;
}

uint32_t rt_event_timedwait(struct rt_event *event, uint32_t wait,
                            unsigned long ticks)
{
    unsigned long start_tick = rt_tick_count();
    uint32_t bits = rt_event_trywait(event, wait);
    if (match_any(bits, wait) || !rt_mutex_timedlock(&event->mutex, ticks))
    {
        return bits;
    }
    while (!match_any(bits = rt_event_trywait(event, wait), wait))
    {
        rt_tick_elapse(&ticks, &start_tick);
        if (!rt_cond_timedwait(&event->cond, &event->mutex, ticks))
        {
            return bits;
        }
    }
    rt_mutex_unlock(&event->mutex);
    return bits;
}

uint32_t rt_event_wait_all(struct rt_event *event, uint32_t wait)
{
    uint32_t bits = rt_event_trywait_all(event, wait);
    if (match_all(bits, wait))
    {
        return bits;
    }
    rt_mutex_lock(&event->mutex);
    while (!match_all(bits = rt_event_trywait_all(event, wait), wait))
    {
        rt_cond_wait(&event->cond, &event->mutex);
    }
    rt_mutex_unlock(&event->mutex);
    return bits;
}

uint32_t rt_event_trywait_all(struct rt_event *event, uint32_t wait)
{
    uint32_t bits = rt_atomic_load(&event->bits, RT_ATOMIC_RELAXED);
    while (match_all(bits, wait) &&
           !rt_atomic_compare_exchange_weak(&event->bits, &bits, bits & ~wait,
                                            RT_ATOMIC_ACQUIRE,
                                            RT_ATOMIC_RELAXED))
    {
    }
    return bits;
}

uint32_t rt_event_timedwait_all(struct rt_event *event, uint32_t wait,
                            unsigned long ticks)
{
    unsigned long start_tick = rt_tick_count();
    uint32_t bits = rt_event_trywait_all(event, wait);
    if (match_all(bits, wait) || !rt_mutex_timedlock(&event->mutex, ticks))
    {
        return bits;
    }
    while (!match_all(bits = rt_event_trywait_all(event, wait), wait))
    {
        rt_tick_elapse(&ticks, &start_tick);
        if (!rt_cond_timedwait(&event->cond, &event->mutex, ticks))
        {
            return bits;
        }
    }
    rt_mutex_unlock(&event->mutex);
    return bits;
}
